﻿using GenLib.Data;
using GenLib.Entity;
using System.Collections.Generic;

namespace GenLib.Builder
{
    /// <summary>
    /// This static class contains utilities for processing directives.
    /// </summary>
    public static class DirectiveUtil
    {
        /// <summary>
        /// This method resolves data object related to variable.
        /// </summary>
        /// <param name="dataRoot">Root of the input data.</param>
        /// <param name="variable">Variable.</param>
        /// <param name="controllingVariables">Symbol table.</param>
        /// <returns>Data object related to variable.</returns>
        public static DataObject ResolveDataObject(DataObject dataRoot, Variable variable, List<Dictionary<string, ControllingVariable>> controllingVariables)
        {
            DataObject dataObject = FindDataObjectInLocalVars(variable, controllingVariables);

            if (dataObject == null)
            {
                dataObject = FindDataObjectInExternalData(variable, dataRoot);
            }

            if (dataObject == null || dataObject.Descendants == null || dataObject.Descendants.Count == 0)
            {
                throw new DataException(ErrorType.DATA_OBJECT_NOT_FOUND);
            }

            return dataObject;
        }

        /// <summary>
        /// This method finds data object related to variable in symbol table.
        /// </summary>
        /// <param name="variable">Variable.</param>
        /// <param name="controllingVariables">Symbol table.</param>
        /// <returns>Data object related to variable.</returns>
        private static DataObject FindDataObjectInLocalVars(Variable variable, List<Dictionary<string, ControllingVariable>> controllingVariables)
        {
            for (int i = controllingVariables.Count - 1; i >= 0; i--)
            {
                Dictionary<string, ControllingVariable> currentControllingVars = controllingVariables[i];

                if (currentControllingVars.ContainsKey(variable.Identificator))
                {
                    return currentControllingVars[variable.Identificator].DataObject;
                }
            }

            return null;
        }

        /// <summary>
        /// This method finds data object related to variable in input data.
        /// </summary>
        /// <param name="variable">Variable.</param>
        /// <param name="dataRoot">Root of input data.</param>
        /// <returns>Data object related to variable.</returns>
        private static DataObject FindDataObjectInExternalData(Variable variable, DataObject dataRoot)
        {
            DataObject currentDataObject = dataRoot;

            if (variable.Identificator.Equals(currentDataObject.Name))
            {
                foreach (var currentProperty in variable.PropertyOrMethodList)
                {

                    if (currentProperty.IsMethod)
                    {
                        return null;
                    }

                    currentDataObject = GetDescendant(currentDataObject.Descendants, currentProperty.Name);

                    if (currentDataObject == null)
                    {
                        return null;
                    }

                }
            }
            else
            {
                return null;
            }

            return currentDataObject;
        }

        /// <summary>
        /// This method get data object related to property from data object list.
        /// </summary>
        /// <param name="dataObjects">Data object.</param>
        /// <param name="propertyName">Name of the property.</param>
        /// <returns>Data object related to property.</returns>
        private static DataObject GetDescendant(List<DataObject> dataObjects, string propertyName)
        {
            if (dataObjects == null)
            {
                return null;
            }

            foreach (var currentDataObject in dataObjects)
            {
                if (currentDataObject.Name.Equals(propertyName))
                {
                    return currentDataObject;
                }
            }

            return null;
        }
    }
}
