﻿using GenLib.Data;
using GenLib.Document;
using GenLib.Entity;
using GenLib.Methods;
using System.Collections.Generic;

namespace GenLib.Builder
{
    /// <summary>
    /// This class is Builder class for Variable entity.
    /// It provides to going through the parsing tree.
    /// </summary>
    class VariableBuilder : Builder
    {
        /// <summary>
        /// This property contains Variable entity.
        /// </summary>
        public Variable Variable { get; set; }

        /// <summary>
        /// Constructor of the VariableBuilder class.
        /// </summary>
        /// <param name="variable">Variable entity.</param>
        public VariableBuilder(Variable variable)
        {
            Variable = variable;
        }

        /// <summary>
        /// This method processes Variable entity. 
        /// It finds variable in symbol table or in input data,
        /// resolves value of the variable,
        /// inserts the value to the final document 
        /// and adds value to the final text of generated document.
        /// </summary>
        public void Build()
        {
            int index;

            string value = FindValueInControllingVarsList(out index);

            if (value == null)
            {
                value = FindValueInData(out index);
            }            

            if (index != -1 && value != null)
            {
                List<PropertyOrMethod> propertyOrMethods = Variable.PropertyOrMethodList;
                List<PropertyOrMethod> subList = propertyOrMethods.GetRange(index, propertyOrMethods.Count - index);
                value = ExecuteMethod(subList, value);
            }

            AddText(value);
            DocumentManager.Instance.InsertValueAtRange(value, Variable.ElementRange);
        }

        /// <summary>
        /// This method finds variable in symbol table and resolve the value of the variable.
        /// </summary>
        /// <param name="index">Index of element in variable input field of document template.</param>
        /// <returns>Value of the variable.</returns>
        private string FindValueInControllingVarsList(out int index)
        {
            index = -1;

            for (int i = ControllingVariables.Count - 1; i >= 0; i--)
            {
                Dictionary<string, ControllingVariable> currentControllingVars = ControllingVariables[i];

                if (currentControllingVars.ContainsKey(Variable.Identificator))
                {
                    return ResolveFromData(currentControllingVars[Variable.Identificator].DataObject, out index);
                }
            }

            return null;
        }

        /// <summary>
        /// This method finds value of the variable in input data.
        /// </summary>
        /// <param name="dataObject">DataObject of input data.</param>
        /// <param name="index">Index of element in variable input field of document template.</param>
        /// <returns>Value of the variable.</returns>
        private string ResolveFromData(DataObject dataObject, out int index)
        {
            index = -1;
            int i = 0;
            string finalValue = null;
            DataObject currentDataObject = dataObject;
            bool endOfTree = false;

            foreach (var currentProperty in Variable.PropertyOrMethodList)
            {
                if (currentProperty.IsMethod)
                {
                    index = i;
                    return finalValue;
                }

                if (endOfTree)
                {
                    return finalValue;
                }
                
                var dataObjectTemp = GetDescendant(currentDataObject.Descendants, currentProperty.Name);

                if (dataObjectTemp == null)
                {
                    finalValue = GetAttribute(currentDataObject.Attributes, currentProperty.Name);

                    if (finalValue == null)
                    {
                        throw new DataException(ErrorType.ATTRIBUTE_DOES_NOT_EXIST);
                    }

                    endOfTree = true;
                }
                else
                {
                    currentDataObject = dataObjectTemp;
                }

                i++;
            }

            return finalValue;
        }

        /// <summary>
        /// This metod finds the variable in attributes list of input data.
        /// </summary>
        /// <param name="attributes">Attributes list.</param>
        /// <param name="name">Name of the variable.</param>
        /// <returns>Value of the variable.</returns>
        private string GetAttribute(List<DataAttribute> attributes, string name)
        {
            if (attributes == null)
            {
                return null;
            }

            foreach (var currentAttribute in attributes)
            {
                if (currentAttribute.Name.Equals(name))
                {
                    return currentAttribute.Value;
                }
            }

            return null;
        }

        /// <summary>
        /// This metod finds the property in data object of input data.
        /// </summary>
        /// <param name="dataObjects">Data object list of input data.</param>
        /// <param name="propertyName">Name of the property.</param>
        /// <returns>Found data object.</returns>
        private DataObject GetDescendant(List<DataObject> dataObjects, string propertyName)
        {
            if (dataObjects == null)
            {
                return null;
            }

            foreach (var currentDataObject in dataObjects)
            {
                if (currentDataObject.Name.Equals(propertyName))
                {
                    return currentDataObject;
                }
            }

            return null;
        }

        /// <summary>
        /// This method finds variable in input data and resolve the value of the variable.
        /// </summary>
        /// <param name="index">Index of element in variable input field of document template.</param>
        /// <returns>Value of the variable.</returns>
        private string FindValueInData(out int index)
        {
            index = -1;

            DataObject currentNode = DataRoot;
            List<PropertyOrMethod> propertyOrMethods = Variable.PropertyOrMethodList;

            if (propertyOrMethods == null || propertyOrMethods.Count <= 0)
            {
                throw new DataException(ErrorType.PROPERTY_DOES_NOT_EXIST);
            }

            if (Variable.Identificator.Equals("this"))
            {
                return ExecuteMethod(propertyOrMethods, null);
            }

            bool correct = false;

            // Processing of the variable
            if (Variable.Identificator.Equals(currentNode.Name))
            {
                if (propertyOrMethods[0].IsMethod)
                {
                    throw new CustomSyntaxException(ErrorType.METHOD_IS_NOT_ALLOWED);
                }

                if (currentNode.Descendants != null && currentNode.Descendants.Count > 0)
                {
                    foreach (var currentDescendant in currentNode.Descendants)
                    {
                        if (currentDescendant.Name.Equals(propertyOrMethods[0].Name))
                        {
                            currentNode = currentDescendant;
                            correct = true;
                            break;
                        }
                    }
                }
                else
                {
                    foreach (var currentAttribute in currentNode.Attributes)
                    {
                        if (currentAttribute.Name.Equals(propertyOrMethods[0].Name))
                        {
                            return currentAttribute.Value;
                        }
                    }

                }
            }
            else
            {
                throw new DataException(ErrorType.VARIABLE_DOES_NOT_EXIST + " Variable: " + Variable.Identificator);
            }

            if (!correct)
            {
                throw new DataException(ErrorType.VARIABLE_DOES_NOT_EXIST + " Variable: " + Variable.Identificator);
            }

            // Processing of the properties and methods
            string finalValue = null;

            for (int i = 0; i < propertyOrMethods.Count - 1; i++)
            {

                if (!correct)
                {
                    break;
                }
                else
                {
                    correct = false;
                }

                if (propertyOrMethods[i].Name.Equals(currentNode.Name))
                {
                    if (currentNode.Descendants != null && currentNode.Descendants.Count > 0)
                    {
                        foreach (var currentDescendant in currentNode.Descendants)
                        {
                            if (currentDescendant.Name.Equals(propertyOrMethods[i + 1].Name))
                            {
                                currentNode = currentDescendant;
                                correct = true;
                                break;
                            }
                        }
                    }
                    else
                    {
                        foreach (var currentAttribute in currentNode.Attributes)
                        {
                            if (currentAttribute.Name.Equals(propertyOrMethods[i + 1].Name))
                            {
                                finalValue = currentAttribute.Value;
                                correct = true;
                                break;
                            }
                        }

                    }

                }
                else
                {
                    if (propertyOrMethods[i + 1].IsMethod)
                    {
                        index = i + 1;
                        return finalValue;
                    }
                    else
                    {
                        throw new DataException(ErrorType.PROPERTY_DOES_NOT_EXIST);
                    }
                }
            }

            return finalValue;

        }

        /// <summary>
        /// This method finds current properties or methods and transform value of the variable according to these.
        /// </summary>
        /// <param name="propertyOrMethods">Property or method list.</param>
        /// <param name="initialInput">Initial value of the variable.</param>
        /// <returns></returns>
        private string ExecuteMethod(List<PropertyOrMethod> propertyOrMethods, string initialInput)
        {
            string input = initialInput;

            foreach(PropertyOrMethod currentMethod in propertyOrMethods)
            {
                if (currentMethod.IsMethod == false)
                {
                    throw new CustomSyntaxException(ErrorType.METHOD_IS_EXPECTED);
                }

                switch(currentMethod.Name)
                {
                    case Constants.CURRENT_DATE:
                        if (input != null)
                        {
                            throw new CustomSyntaxException(ErrorType.METHOD_IS_EXPECTED_CURRENT_DATE);
                        }

                        input = MethodsLib.MethodCurrentDate();
                        break;

                    case Constants.FORMAT_DATE:
                        if (currentMethod.Parameter == null || input == null)
                        {
                            throw new CustomSyntaxException(ErrorType.METHOD_IS_EXPECTED_FORMAT_DATE);
                        }

                        input = MethodsLib.MethodFormatDate(input, currentMethod.Parameter);
                        break;
                }
            }

            return input;
        }
    }
}
