﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using Word = Microsoft.Office.Interop.Word;

namespace GenLib.Document
{
    /// <summary>
    /// This static class contains utilities for working with Word files.
    /// </summary>
    public static class DocumentUtil
    {
        /// <summary>
        /// List of running Word processes started by this program.
        /// </summary>
        public static List<Process> currentWordProcesses = new List<Process>();

        /// <summary>
        /// This method loads Word document from file.
        /// </summary>
        /// <param name="filePath">Word file path.</param>
        /// <param name="isReadOnly">Flag, true - if Word document will be opened in read only mode, false - otherwise.</param>
        /// <returns>Instance of opened Word document.</returns>
        public static Word.Document LoadWordFile(string filePath, bool isReadOnly)
        {
            if (!File.Exists(filePath))
            {
                throw new FileNotFoundException(ErrorType.FILE_NOT_FOUND + "File path: " + filePath);
            }

            Process[] processesBefore = GetWordProcesses();
            Word.Application wordApp = new Word.Application();
            Process[] processesAfter = GetWordProcesses();
            GetCurrentProcess(processesBefore, processesAfter);

            Word.Document wordFile;

            wordApp.Visible = false;

            object path = filePath;
            object missing = Type.Missing;
            object isReadOnlyObj = isReadOnly;

            wordFile = wordApp.Documents.Open(ref path,
                ref missing, ref isReadOnlyObj, ref missing, ref missing, ref missing,
                ref missing, ref missing, ref missing, ref missing, ref missing,
                ref missing, ref missing, ref missing, ref missing, ref missing);

            return wordFile;
        }

        /// <summary>
        /// This method gets currently opened Word process during loading Word file.
        /// </summary>
        /// <param name="processesBefore">Array of running Word processes before loading Word file.</param>
        /// <param name="processesAfter">Array of running Word processes after loading Word file.</param>
        private static void GetCurrentProcess(Process[] processesBefore, Process[] processesAfter)
        {
            if (processesAfter.Length < processesBefore.Length)
            {
                return;
            }

            bool found;

            for (int i = 0; i < processesAfter.Length; i++)
            {
                found = false;

                for (int j = 0; j < processesBefore.Length; j++)
                {
                    if (processesBefore[j].Id == processesAfter[i].Id)
                    {
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    currentWordProcesses.Add(processesAfter[i]);
                }
            }
        }

        /// <summary>
        /// This method get currently running Word processes in Task Manager.
        /// </summary>
        /// <returns>Currently running Word processes.</returns>
        private static Process[] GetWordProcesses()
        {
            Process[] wordProcesses = Process.GetProcessesByName("WINWORD");
            return wordProcesses;
        }
    }
}
