﻿using GenLib.Data;
using GenLib.Document.Entity;
using GenLib.Entity;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using Word = Microsoft.Office.Interop.Word;

namespace GenLib.Document
{
    /// <summary>
    /// Main class of the GenLib library project. It provides to load document template, 
    /// load input data, preprocess the template and generate and save final Word document.
    /// This class is according to design pattern Singleton.
    /// </summary>
    public class DocumentManager
    {
        /// <summary>
        /// Constructor of the DocumentManager class.
        /// </summary>
        private DocumentManager() { }

        /// <summary>
        /// This attribute contains instance of the DocumentManager class.
        /// </summary>
        private static readonly DocumentManager INSTANCE = new DocumentManager();

        /// <summary>
        /// Property of the INSTANCE attribute.
        /// </summary>
        public static DocumentManager Instance
        {
            get { return INSTANCE; }
        }

        /// <summary>
        /// Property of the path to temporary document template.
        /// </summary>
        private string TempTemplatePath { get; set; }

        /// <summary>
        /// Property of the path to generated Word document.
        /// </summary>
        private string DocumentPath { get; set; }

        /// <summary>
        /// Property of the instance of temporary document template.
        /// </summary>
        private Word.Document TempTemplate { get; set; }

        /// <summary>
        /// Property of the instance of generated Word document.
        /// </summary>
        private Word.Document Document { get; set; }

        /// <summary>
        /// This method loads temporary document template from file.
        /// </summary>
        /// <param name="templatePath">Path to the temporary document template.</param>
        public void LoadTempTemplateFromFile(string templatePath)
        {
            string tempTemplatePath = CreateTempTemplate();

            CopyTemplateContentToTempTemplate(templatePath, tempTemplatePath);
        }

        /// <summary>
        /// This method loads temporary document template.
        /// </summary>
        /// <param name="template">Instance of the document template.</param>
        public void LoadTempTemplate(Word.Document template)
        {
            string tempTemplatePath = CreateTempTemplate();

            CopyTemplateContentToTempTemplate(template, tempTemplatePath);
        }

        /// <summary>
        /// This method copies content of the document template file 
        /// to the temporary document template file.
        /// </summary>
        /// <param name="source">Document template file path.</param>
        /// <param name="destination">Temporary document template file path.</param>
        private void CopyTemplateContentToTempTemplate(string source, string destination)
        {
            File.Copy(source, destination, true);

            TempTemplate = DocumentUtil.LoadWordFile(destination, false);
            TempTemplatePath = destination;
        }

        /// <summary>
        /// This method copies content of the document template 
        /// to the temporary document template file.
        /// </summary>
        /// <param name="template">Instance of open document template.</param>
        /// <param name="destination">Temporary document template file path.</param>
        public void CopyTemplateContentToTempTemplate(Word.Document template, string destination)
        {
            TempTemplate = DocumentUtil.LoadWordFile(destination, false);
            TempTemplatePath = destination;

            Word.Range templateContent = template.Content;
            Word.Range tempTemplateContent = TempTemplate.Range(TempTemplate.Content.End - 1, TempTemplate.Content.End - 1);

            templateContent.Copy();
            tempTemplateContent.PasteAndFormat(Word.WdRecoveryType.wdFormatOriginalFormatting);

            TempTemplate.Range(TempTemplate.Content.End - 1, TempTemplate.Content.End).Delete();

            TempTemplate.Save();
        }

        /// <summary>
        /// This method creates temporary document template file.
        /// </summary>
        /// <returns>Temporary document template file path.</returns>
        private string CreateTempTemplate()
        {
            string tempFilePath = string.Empty;
            tempFilePath = Path.GetTempFileName();

            return tempFilePath;
        }

        /// <summary>
        /// This method provides to preprocess the document template.
        /// </summary>
        public void Preprocess()
        {
            DocumentPreprocessor documentPreprocessor = new DocumentPreprocessor(TempTemplate);
            documentPreprocessor.Preprocess();
            TempTemplate.Save();
        }

        /// <summary>
        /// This method creates final Word document for the generation.
        /// </summary>
        /// <param name="documentPath">New Word document file path.</param>
        public void CreateDocument(string documentPath)
        {
            File.Copy(TempTemplatePath, documentPath, true);

            Document = DocumentUtil.LoadWordFile(documentPath, false);
            DocumentPath = documentPath;
        }

        /// <summary>
        /// This method provides to generate final Word document via input data.
        /// </summary>
        /// <param name="dataFile">Input data file path.</param>
        public void GenerateDocument(string dataFile)
        {
            DataLoader loader = new DataLoader(dataFile);
            loader.LoadDataJson();
            string documentText = Document.Content.Text;

            documentText = RemoveTableCharsFromText(documentText);

            Generator.Instance.Generate(documentText, loader.Root);
        }

        /// <summary>
        /// This method removes table characters from the text (for needs of tableforeach directive).
        /// </summary>
        /// <param name="text">Input text containing table characters.</param>
        /// <returns>Text without the table characters.</returns>
        private string RemoveTableCharsFromText(string text)
        {
            return text.Replace("\a", "");
        }

        /// <summary>
        /// This method provides to analyze in-process document template (for need of TemplateBuilder).
        /// </summary>
        /// <param name="template">Instance of open document template.</param>
        /// <param name="currentIndex">Index of current cursor position in opened document template.</param>
        /// <returns>Parameters of the document template in a relation to current cursor position in opened document template.</returns>
        public DocumentParams AnalyzeDocument(Word.Document template, int currentIndex)
        {
            Generator generator = Generator.Instance;
            string templateText = template.Content.Text;

            templateText = RemoveTableCharsFromText(templateText);

            return SearchGrammarObjectByIndex(generator.Analyze(templateText), currentIndex);
        }

        /// <summary>
        /// This method searches language construction in which is the position of the cursor in opened document template.
        /// </summary>
        /// <param name="parse">Parse entity of the parsing tree.</param>
        /// <param name="index">Index of current cursor position in opened document template.</param>
        /// <returns>Parameters of the document template in a relation to current cursor position in opened document template.</returns>
        private DocumentParams SearchGrammarObjectByIndex(Parse parse, int index)
        {
            ForeachDirective foundForeachDirective = SearchForeachDirectiveByIndex(parse.Block, index, null);
            TableForeachDirective foundTableForeachDirective = SearchTableForeachDirectiveByIndex(parse.Block, index, null);

            if (foundForeachDirective != null)
            {
                return new DocumentParamsForeachDirective(foundForeachDirective.ControlVariableId, foundForeachDirective.ControlArray);
            }
            else if (foundTableForeachDirective != null)
            {
                return new DocumentParamsTableForeachDirective(foundTableForeachDirective.ControlVariableId, foundTableForeachDirective.ControlArray);
            }

            return null;
        }

        /// <summary>
        /// This method recursively searches through all foreach directives if the current position of the cursor is in it.
        /// </summary>
        /// <param name="block">Block entity of the parsing tree.</param>
        /// <param name="index">Index of current cursor position in opened document template.</param>
        /// <param name="foundForeachDirective">Founded foreach directive.</param>
        /// <returns>Founded foreach directive.</returns>
        private ForeachDirective SearchForeachDirectiveByIndex(Block block, int index, ForeachDirective foundForeachDirective)
        {
            List<Atom> atoms = block.Atoms;

            foreach (Atom currentAtom in atoms)
            {
                if (currentAtom is ForeachDirective currentForeachDirective)
                {
                    Block currentBlock = currentForeachDirective.Block;
                    ElementRange curentElementRange = currentBlock.ElementRange;

                    int blockStart = curentElementRange.Start;
                    int blockStop = curentElementRange.Stop;

                    if (index >= blockStart && index <= blockStop)
                    {
                        return SearchForeachDirectiveByIndex(currentBlock, index, currentForeachDirective);
                    }
                }
            }

            return foundForeachDirective;
        }

        /// <summary>
        /// This method recursively searches through all tableforeach directives if the current position of the cursor is in it.
        /// </summary>
        /// <param name="block">Block entity of the parsing tree.</param>
        /// <param name="index">Index of current cursor position in opened document template.</param>
        /// <param name="foundTableForeachDirective">Founded tableforeach directive.</param>
        /// <returns>Founded tableforeach directive.</returns>
        private TableForeachDirective SearchTableForeachDirectiveByIndex(Block block, int index, TableForeachDirective foundTableForeachDirective)
        {
            List<Atom> atoms = block.Atoms;

            foreach (Atom currentAtom in atoms)
            {
                if (currentAtom is TableForeachDirective currentTableForeachDirective)
                {
                    Block currentBlock = currentTableForeachDirective.Block;
                    ElementRange curentElementRange = currentBlock.ElementRange;

                    int blockStart = curentElementRange.Start;
                    int blockStop = curentElementRange.Stop;

                    if (index >= blockStart && index <= blockStop)
                    {
                        return SearchTableForeachDirectiveByIndex(currentBlock, index, currentTableForeachDirective);
                    }
                }
            }

            return foundTableForeachDirective;
        }

        /// <summary>
        /// This method closes opened temporary document template and final Word document.
        /// </summary>
        public void Close()
        {
            if (Document != null)
            {
                Document.Save();
                Document.Close();
                Document = null;
            }

            if (TempTemplate != null)
            {
                TempTemplate.Save();
                TempTemplate.Close();
                TempTemplate = null;
            }

            if (File.Exists(TempTemplatePath))
            {
                File.Delete(TempTemplatePath);
                TempTemplatePath = null;
            }
        }

        /// <summary>
        /// This method closes all opened Word processes started by this program.
        /// </summary>
        public void CloseWordProcesses()
        {
            List<Process> wordProcesses = DocumentUtil.currentWordProcesses;

            for (int i = 0; i < wordProcesses.Count; i++)
            {
                Process currentProcess = wordProcesses[i];

                if (currentProcess != null || !currentProcess.HasExited)
                {
                    currentProcess.Kill();
                }
            }

            wordProcesses.Clear();            
        }

        /// <summary>
        /// This method copies specified range in Word document and insert after input index in Word document.
        /// </summary>
        /// <param name="rangeToCopy">Range to copy.</param>
        /// <param name="insertAfter">Index to insert copied range after it.</param>
        public void CopyRange(Word.Range rangeToCopy, int insertAfter)
        {
            if (IsFirstCharWhiteSpace(rangeToCopy))
            {
                rangeToCopy.Start += 1;
            }

            Word.Range targetRange = rangeToCopy.Duplicate;

            targetRange.Start = insertAfter;

            targetRange.FormattedText = rangeToCopy.FormattedText;
         }

        /// <summary>
        /// This method copies specified range with table in Word document and insert after input index in Word document.
        /// </summary>
        /// <param name="rangeToCopy">Range with table to copy.</param>
        /// <param name="insertAfter">Index to insert copied range after it.</param>
        public void CopyTable(Word.Range rangeToCopy, int insertAfter)
        {
            Word.Range targetRange = rangeToCopy.Duplicate;

            targetRange.Start = insertAfter;

            targetRange.FormattedText = rangeToCopy.FormattedText;
        }

        /// <summary>
        /// This method detecs if first character of input range is whitespace.
        /// </summary>
        /// <param name="range">Range in Word document.</param>
        /// <returns>true - if first character of the input range is whitespace; false - otherwise.</returns>
        private bool IsFirstCharWhiteSpace(Word.Range range)
        {
            char firstChar = range.Text[0];

            if (Char.IsWhiteSpace(firstChar))
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// This method copies specified range from document template to generated Word document
        /// and insert after input index in Word document.
        /// </summary>
        /// <param name="rangeToCopy">Range to copy.</param>
        /// <param name="insertAfter">Index to insert copied range after it.</param>
        public void CopyRangeFromTemplate(ElementRange rangeToCopy, int insertAfter)
        {
            Word.Range templateRange = TempTemplate.Range(rangeToCopy.Start, rangeToCopy.Stop + 1);
            Word.Range documentRange = Document.Range(insertAfter, insertAfter);

            templateRange.Copy();
            documentRange.Paste();
        }

        /// <summary>
        /// This method deletes range specified range in Word document.
        /// </summary>
        /// <param name="rangeToDelete">Range to delete.</param>
        public void DeleteRange(Word.Range rangeToDelete)
        {
            rangeToDelete.Delete();
        }

        /// <summary>
        /// This method convert element range to an instance of Word Range class.
        /// </summary>
        /// <param name="elementRange">Range of the element to convert.</param>
        public void AddWordRange(ElementRange elementRange)
        {
            if (Document != null)
            {
                Word.Range wordRange = Document.Range(elementRange.Start, elementRange.Stop + 1);
                elementRange.Range = wordRange;
            }
        }

        /// <summary>
        /// This method inserts new row into the table within the specified Word Range.
        /// </summary>
        /// <param name="range">Range with table.</param>
        /// <param name="requiredTablesCount">Required tables count in specified range.</param>
        public void AddRowIntoTableAfterHeader(Word.Range range, int requiredTablesCount)
        {
            int tablesCount = range.Tables.Count;

            if (tablesCount != requiredTablesCount)
            {
                throw new CustomSyntaxException(ErrorType.INCONSISTENT_TABLE);
            }
            object beforeRow = range.Tables[1].Rows[2];

            range.Tables[1].Rows.Add(beforeRow);
        }

        /// <summary>
        /// This method copies initial table values from document template.
        /// </summary>
        /// <param name="sourceElementRange">Element range in document template.</param>
        /// <param name="targetRange">Range with the table to copy values.</param>
        public void CopyTableValuesFromTemplate(ElementRange sourceElementRange, Word.Range targetRange)
        {
            Word.Range sourceRange = TempTemplate.Range(sourceElementRange.Start, sourceElementRange.Stop + 1);

            int sourceTablesCount = sourceRange.Tables.Count;
            int targetTableCount = targetRange.Tables.Count;

            if (sourceTablesCount != 1 || targetTableCount != 1)
            {
                throw new CustomSyntaxException(ErrorType.INCONSISTENT_TABLE);
            }

            Word.Table sourceTable = sourceRange.Tables[1];
            Word.Table targetTable = targetRange.Tables[1];

            if (sourceTable.Rows.Count != 2)
            {
                throw new CustomSyntaxException(ErrorType.INCONSISTENT_TABLE);
            }

            if (targetTable.Rows.Count < 2)
            {
                throw new CustomSyntaxException(ErrorType.INCONSISTENT_TABLE);
            }

            for (int i = 1; i <= sourceTable.Columns.Count; i++)
            {
                Word.Range currentSourceRange = sourceTable.Cell(2, i).Range;
                Word.Range currentTargetRange = targetTable.Cell(2, i).Range;

                object oCharacter = Word.WdUnits.wdCharacter;
                object negOne = -1;

                currentSourceRange.MoveEnd(oCharacter, negOne);
                currentTargetRange.MoveEnd(oCharacter, negOne);

                currentSourceRange.Copy();
                currentTargetRange.PasteAndFormat(Word.WdRecoveryType.wdFormatOriginalFormatting);
            }

        }

        /// <summary>
        /// This method inserts specified value at the specified range in Word document.
        /// </summary>
        /// <param name="value">Value to insert.</param>
        /// <param name="range">Range in Word document.</param>
        public void InsertValueAtRange(string value, ElementRange range)
        {
            if (Document != null)
            {
                Word.Range wordRange = Document.Range(range.Start, range.Stop + 1);
                wordRange.Text = value;
            }
        }
    }
}
