﻿using DocsGen.Services;
using GenLib;
using GenLib.Document;
using GenLib.Document.Entity;
using System;
using System.Drawing;
using System.Windows.Forms;
using TemplateBuilder.Model.Entities;
using TemplateBuilder.Model.Entities.Metadata;
using TemplateBuilder.Model.Entities.NodeObjects;
using TemplateBuilder.Model.Entities.NodeObjects.Directives;
using TemplateBuilder.Model.Entities.NodeObjects.Methods;
using TemplateBuilder.Model.Services;
using TemplateBuilder.Model.Utils;
using Word = Microsoft.Office.Interop.Word;

namespace TemplateBuilder.Forms
{
    /// <summary>
    /// This partial class serves actions for TemplateTaskPane. 
    /// The form mediates GUI for the view of metadata tree.
    /// </summary>
    public partial class TemplateTaskPane : UserControl
    {
        /// <summary>
        /// Label with text 'Node:'.
        /// </summary>
        private Label nodeTextLabel;

        /// <summary>
        /// Name of the attribute or object.
        /// </summary>
        private Label nodeValueLabel;

        /// <summary>
        /// Add button for insert language construction into document template.
        /// </summary>
        private Button addButton;

        /// <summary>
        /// Label with text 'Directives:'.
        /// </summary>
        private Label directivesLabel;

        /// <summary>
        /// Combo box with available directives for object in metadata tree view.
        /// </summary>
        private ComboBox directivesComboBox;

        /// <summary>
        /// Label with text 'Methods'.
        /// </summary>
        private Label methodsLabel;

        /// <summary>
        /// Check box for Format Date method.
        /// </summary>
        private CheckBox formatDateMethodCheckBox;

        /// <summary>
        /// Text box for Format Date method parameters.
        /// </summary>
        private TextBox formatDateMethodParamsTextBox;

        /// <summary>
        /// Label with text 'Controlling variable' for foreach directive.
        /// </summary>
        private Label foreachDirectiveControlVariableLabel;

        /// <summary>
        /// Name of the controlling variable for foreach.
        /// </summary>
        private TextBox foreachDirectiveParamsTextBox;

        /// <summary>
        /// Label with text 'Controlling variable' for tableforeach directive.
        /// </summary>
        private Label tableForeachDirectiveControlVariableLabel;

        /// <summary>
        /// Name of the controlling variable for tableforeach.
        /// </summary>
        private TextBox tableForeachDirectiveParamsTextBox;

        /// <summary>
        /// Index of last curson position in document template.
        /// </summary>
        private int lastStartIndex = -1;

        /// <summary>
        /// Instance of TreeViewManager class.
        /// </summary>
        private TreeViewManager treeViewManager;

        /// <summary>
        /// Constructor of the TemplateTaskPane class.
        /// </summary>
        public TemplateTaskPane()
        {
            InitializeComponent();
            InitializeCustomComponent();
        }

        /// <summary>
        /// This method initializes custom components of the TemplateTaskPane.
        /// </summary>
        private void InitializeCustomComponent()
        {
            refreshButton.Image = ImageConvertor.Convert(Globals.ThisAddIn.Application.CommandBars.GetImageMso("Refresh", 18, 18));

            nodeTextLabel = new Label
            {
                Size = new Size(34, 20),
                Location = new Point(6, 25),
                Text = "Node:"
            };

            nodeValueLabel = new Label
            {
                Size = new Size(121, 20),
                Location = new Point(45, 25)
            };

            directivesLabel = new Label
            {
                Size = new Size(60, 20),
                Location = new Point(6, 45),
                Text = "Directives"
            };

            directivesComboBox = new ComboBox
            {
                Size = new Size(121, 20),
                Location = new Point(20, 68)
            };

            methodsLabel = new Label
            {
                Size = new Size(60, 20),
                Location = new Point(6, 45),
                Text = "Methods"
            };

            //FORMAT DATE METHOD
            formatDateMethodCheckBox = new CheckBox
            {
                Size = new Size(86, 20),
                Location = new Point(20, 68),
                CheckState = CheckState.Unchecked,
                Text = "Format Date"
            };

            formatDateMethodParamsTextBox = new TextBox
            {
                Size = new Size(184, 20),
                Location = new Point(110, 66),
                Enabled = false,
                Text = "dd.MM.yyyy"
            };

            formatDateMethodCheckBox.CheckedChanged += (s, e) => FormatDateMethodCheckBox_CheckedChanged(s, e);

            //FOREACH DIRECTIVE
            foreachDirectiveControlVariableLabel = new Label
            {
                Size = new Size(120, 20),
                Location = new Point(20, 98),
                Text = "Controlling variable:"
            };

            foreachDirectiveParamsTextBox = new TextBox
            {
                Size = new Size(100, 46),
                Location = new Point(22, 120),
                MaxLength = 16,
                Text = "variable"
            };

            //TABLE FOREACH DIRECTIVE
            tableForeachDirectiveControlVariableLabel = new Label
            {
                Size = new Size(120, 20),
                Location = new Point(20, 98),
                Text = "Controlling variable:"
            };

            tableForeachDirectiveParamsTextBox = new TextBox
            {
                Size = new Size(100, 46),
                Location = new Point(22, 120),
                MaxLength = 16,
                Text = "variable"
            };

            directivesComboBox.SelectedIndexChanged += (s, e) => DirectivesComboBox_SelectedIndexChanged(s, e);
        }

        /// <summary>
        /// This method serves action checked format date check box changed in TemplateTaskPane.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void FormatDateMethodCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            CheckBox checkBox = sender as CheckBox;

            if (checkBox.Checked)
            {
                formatDateMethodParamsTextBox.Enabled = true;
            }
            else
            {
                formatDateMethodParamsTextBox.Enabled = false;
            }
        }

        /// <summary>
        /// This method serves action selected index of directives combo box changed in TemplateTaskPane.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void DirectivesComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            ComboBox comboBox = sender as ComboBox;

            //CLEAN
            detailsGroupBox.Controls.Remove(foreachDirectiveControlVariableLabel);
            detailsGroupBox.Controls.Remove(foreachDirectiveParamsTextBox);

            detailsGroupBox.Controls.Remove(tableForeachDirectiveControlVariableLabel);
            detailsGroupBox.Controls.Remove(tableForeachDirectiveParamsTextBox);

            if (addButton != null)
            {
                addButton.Enabled = false;
            }

            //ADD CORRECT ARGS
            if (comboBox.SelectedIndex > -1)
            {
                if (comboBox.SelectedItem.Equals(Directives.FOREACH))
                {
                    DirectivesComboBoxShowForeach(comboBox);
                }
                else if (comboBox.SelectedItem.Equals(Directives.TABLE_FOREACH))
                {
                    DirectivesComboBoxShowTableForeach(comboBox);
                }
            }            
        }

        /// <summary>
        /// This method adds foreach directive to directives combo box.
        /// </summary>
        /// <param name="comboBox">Directives combo box.</param>
        private void DirectivesComboBoxShowForeach(ComboBox comboBox)
        {
            detailsGroupBox.Controls.Add(foreachDirectiveControlVariableLabel);
            detailsGroupBox.Controls.Add(foreachDirectiveParamsTextBox);
            addButton.Enabled = true;
        }

        /// <summary>
        /// This method adds tableforeach directive to directives combo box.
        /// </summary>
        /// <param name="comboBox">Directives combo box.</param>
        private void DirectivesComboBoxShowTableForeach(ComboBox comboBox)
        {
            detailsGroupBox.Controls.Add(tableForeachDirectiveControlVariableLabel);
            detailsGroupBox.Controls.Add(tableForeachDirectiveParamsTextBox);
            addButton.Enabled = true;
        }

        /// <summary>
        /// This method provides to initialize Metadata Tree View.
        /// </summary>
        /// <param name="root">Metadata root.</param>
        public void TreeViewInit(MetadataObject root)
        {
            TreeViewInit(root, null, null);
        }

        /// <summary>
        /// This method provides to initialize Metadata Tree View.
        /// </summary>
        /// <param name="root">Metadata root.</param>
        /// <param name="controlVariableId">Identifier of controlling variable for foreach and tableforeach directives.</param>
        /// <param name="controlArray">Identifier of controlling array for foreach and tableforeach directives.</param>
        public void TreeViewInit(MetadataObject root, string controlVariableId, string controlArray)
        {
            treeViewManager = new TreeViewManager(treeView);

            treeViewManager.UpdateTreeView(root, controlVariableId, controlArray);
        }

        /// <summary>
        /// This method serves action on tree view node click.
        /// It provides to show node name and node info in details section in TemplateTaskPane.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void TreeView_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            try
            {
                timer1.Enabled = true;
                Globals.Ribbons.TemplateRibbon.showErrorsButton.Enabled = true;
                Globals.Ribbons.TemplateRibbon.collapseErrorsButton.Enabled = true;

                NodeManager nodeManager = new NodeManager(e.Node);

                detailsGroupBox.Controls.Clear();

                ShowNodeName(nodeManager);
                ShowNodeInfo(nodeManager);
            }
            catch(Exception ex)
            {
                string infoMessage = "Metadata tree view error.\nError:\n" + ex.Message;
                MessageBox.Show(infoMessage, "Error: Metadata Tree View", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// This method shows node name in details section in TemplateTaskPane.
        /// </summary>
        /// <param name="nodeManager">Instance of current NodeManager.</param>
        private void ShowNodeName(NodeManager nodeManager)
        {
            nodeValueLabel.Text = nodeManager.Node.Name;

            detailsGroupBox.Controls.Add(nodeTextLabel);
            detailsGroupBox.Controls.Add(nodeValueLabel);
        }

        /// <summary>
        /// This method shows node info in details section in TemplateTaskPane.
        /// </summary>
        /// <param name="nodeManager">Instance of current NodeManager.</param>
        private void ShowNodeInfo(NodeManager nodeManager)
        {
            if (nodeManager.NodeObject.ObjectType.Equals(ENodeObjectType.Attribute))
            {
                ShowAttributeInfo(nodeManager);
            }
            else
            {
                ShowElementInfo(nodeManager);
            }                        
        }

        /// <summary>
        /// This method shows attribute info in details section in TemplateTaskPane.
        /// </summary>
        /// <param name="nodeManager">Instance of current NodeManager.</param>
        private void ShowAttributeInfo(NodeManager nodeManager)
        {
            NodeObjectAttribute attribute = (NodeObjectAttribute)nodeManager.NodeObject;
            
            GenerateMethodsSection(attribute);
            ShowAddButton();
            addButton.Enabled = true;
            addButton.Click += (s, e) => AddButtonAttribute_Click(s, e, attribute, nodeManager);
        }

        /// <summary>
        /// This method shows element info in details section in TemplateTaskPane.
        /// </summary>
        /// <param name="nodeManager">Instance of current NodeManager.</param>
        private void ShowElementInfo(NodeManager nodeManager)
        {
            NodeObjectElement element = (NodeObjectElement)nodeManager.NodeObject;

            if (element.IsArray)
            {
                ShowAddButton();
                GenerateDirectivesSection();
                
                addButton.Click += (s, e) => AddButtonElement_Click(s, e, element, nodeManager);
            }
        }

        /// <summary>
        /// This method show add button in details section in TemplateTaskPane.
        /// </summary>
        private void ShowAddButton()
        {
            addButton = new Button
            {
                Size = new Size(75, 23),
                Location = new Point(233, 184),
                UseVisualStyleBackColor = true,
                Enabled = false,
                Text = "Add"
            };

            detailsGroupBox.Controls.Add(addButton);
        }

        /// <summary>
        /// This method serves action on add button attribute click.
        /// It provides to add language construction into the document template.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        /// <param name="attribute">Current attribute.</param>
        /// <param name="nodeManager">Instance of current NodeManager.</param>
        private void AddButtonAttribute_Click(object sender, EventArgs e, NodeObjectAttribute attribute, NodeManager nodeManager)
        {
            try
            {
                if (detailsGroupBox.Contains(formatDateMethodCheckBox))
                {
                    if (formatDateMethodCheckBox.Checked)
                    {
                        attribute.AddMethod(ENodeObjectMethodType.FormatDate, formatDateMethodParamsTextBox.Text);
                    }
                    else
                    {
                        attribute.RemoveMethod(ENodeObjectMethodType.FormatDate);
                    }
                }

                WordManager.InsertMark(nodeManager.TemplateMark);
            }
            catch (Exception ex)
            {
                string infoMessage = "Adding Format Date method failed.\nError:\n" + ex.Message;
                MessageBox.Show(infoMessage, "Error: Format Date method", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// This method serves action on add button element click.
        /// It provides to add language construction into the document template.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        /// <param name="element">Current attribute.</param>
        /// <param name="nodeManager">Instance of current NodeManager.</param>
        private void AddButtonElement_Click(object sender, EventArgs e, NodeObjectElement element, NodeManager nodeManager)
        {
            try
            {
                if (directivesComboBox.SelectedItem.Equals(Directives.FOREACH))
                {
                    element.AddDirective(ENodeObjectDirectiveType.Foreach, foreachDirectiveParamsTextBox.Text);
                    WordManager.InsertMark(nodeManager.TemplateMark);
                }
                else if (directivesComboBox.SelectedItem.Equals(Directives.TABLE_FOREACH))
                {
                    element.AddDirective(ENodeObjectDirectiveType.TableForeach, tableForeachDirectiveParamsTextBox.Text);
                    OpenForeachTableCreator(nodeManager);
                }
            }
            catch(Exception ex)
            {
                string infoMessage = "Adding directive failed.\nError:\n" + ex.Message;
                MessageBox.Show(infoMessage, "Error: Adding directive", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// This method shows TableForeachForm.
        /// </summary>
        /// <param name="nodeManager">Instance of current NodeManager.</param>
        private void OpenForeachTableCreator(NodeManager nodeManager)
        {
            addButton.Enabled = false;

            TableForeachForm tableForeachForm = new TableForeachForm(nodeManager);
            tableForeachForm.Show();
            tableForeachForm.FormClosed += (s, e) => { addButton.Enabled = true; };
        }

        /// <summary>
        /// This method generates section for directives in Details section.
        /// </summary>
        private void GenerateDirectivesSection()
        {
            detailsGroupBox.Controls.Add(directivesLabel);
            detailsGroupBox.Controls.Add(directivesComboBox);

            AddForeachDirective(directivesComboBox);
            AddTableForeachDirective(directivesComboBox);
        }

        /// <summary>
        /// This method provides to add foreach directive to directives combo box.
        /// </summary>
        /// <param name="directivesComboBox">Directives combo box.</param>
        private void AddForeachDirective(ComboBox directivesComboBox)
        {
            if (!directivesComboBox.Items.Contains(Directives.FOREACH))
            {
                directivesComboBox.Items.Add(Directives.FOREACH);
            }

            if (directivesComboBox.SelectedItem != null && directivesComboBox.SelectedItem.Equals(Directives.FOREACH))
            {
                DirectivesComboBoxShowForeach(directivesComboBox);
            }
        }

        /// <summary>
        /// This method provides to add tableforeach directive to directives combo box.
        /// </summary>
        /// <param name="directivesComboBox">Directives combo box.</param>
        private void AddTableForeachDirective(ComboBox directivesComboBox)
        {
            if (!directivesComboBox.Items.Contains(Directives.TABLE_FOREACH))
            {
                directivesComboBox.Items.Add(Directives.TABLE_FOREACH);
            }

            if (directivesComboBox.SelectedItem != null && directivesComboBox.SelectedItem.Equals(Directives.TABLE_FOREACH))
            {
                DirectivesComboBoxShowTableForeach(directivesComboBox);
            }
        }

        /// <summary>
        /// This method generates section for methods in Details section.
        /// </summary>
        /// <param name="attribute">Current attribute.</param>
        private void GenerateMethodsSection(NodeObjectAttribute attribute)
        {
            detailsGroupBox.Controls.Add(methodsLabel);

            if (attribute.DataType.Equals(EDataType.DateTime))
            {
                AddFormatDateMethod(attribute);
            }            
        }

        /// <summary>
        /// This method adds Format Date method in the Details section.
        /// </summary>
        /// <param name="attribute">Current attribute.</param>
        private void AddFormatDateMethod(NodeObjectAttribute attribute)
        {
            detailsGroupBox.Controls.Add(formatDateMethodCheckBox);
            detailsGroupBox.Controls.Add(formatDateMethodParamsTextBox);
        }

        /// <summary>
        /// This method serves action on refresh button click.
        /// It provides to refresh metadata tree view according to current cursor position
        /// in document template.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void RefreshButton_Click(object sender, EventArgs e)
        {
            if (Globals.ThisAddIn.MetadataLoader.Root == null)
            {
                return;
            }

            Word.Application currentTemplate = Globals.ThisAddIn.Application;
            int startIndex = currentTemplate.Selection.Range.Start;

            DocumentManager documentManager = DocumentManager.Instance;
            DocumentParams documentParams;

            try
            {
                documentParams = documentManager.AnalyzeDocument(currentTemplate.ActiveDocument, startIndex);
                Globals.ThisAddIn.ErrorTaskPane.RemoveErrors();
                RefreshTreeView(documentParams);
            }
            catch(AntlrSyntaxException asx)
            {
                Globals.ThisAddIn.ErrorTaskPane.AddError(asx.Message);
            }
            catch(CustomSyntaxException csx)
            {
                Globals.ThisAddIn.ErrorTaskPane.AddError(csx.Message);
            }
        }

        /// <summary>
        /// This method refreshes metadata tree view according to current cursor position
        /// in document template.
        /// </summary>
        /// <param name="documentParams">Parameters of current document template.</param>
        private void RefreshTreeView(DocumentParams documentParams)
        {
            MetadataObject root = Globals.ThisAddIn.MetadataLoader.Root;
            if (documentParams == null)
            {
                if (treeViewManager.IsSameTree(root))
                {
                    return;
                }

                TreeViewInit(root);
            }

            if (documentParams is DocumentParamsForeachDirective documentParamsForeachDirective)
            {
                MetadataObject currentNode = MetadataExecutor.FindElementByPath(root, documentParamsForeachDirective.ControlArray.Split('.'));

                if (currentNode.Descendants == null || currentNode.Descendants.Count <= 0)
                {
                    return;
                }

                currentNode = currentNode.Descendants[0];

                if (treeViewManager.IsSameTree(currentNode))
                {
                    return;
                }

                TreeViewInit(currentNode,
                    documentParamsForeachDirective.ControlVariableId,
                    documentParamsForeachDirective.ControlArray
                );
            }
            else if (documentParams is DocumentParamsTableForeachDirective documentParamstableForeachDirective)
            {
                MetadataObject currentNode = MetadataExecutor.FindElementByPath(root, documentParamstableForeachDirective.ControlArray.Split('.'));

                if (currentNode.Descendants == null || currentNode.Descendants.Count <= 0)
                {
                    return;
                }

                currentNode = currentNode.Descendants[0];

                if (treeViewManager.IsSameTree(currentNode))
                {
                    return;
                }

                TreeViewInit(currentNode,
                    documentParamstableForeachDirective.ControlVariableId,
                    documentParamstableForeachDirective.ControlArray
                );
            }
        }

        /// <summary>
        /// This method serves action for tick of the timer. 
        /// It provides to check cursor position changed.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void Timer1_Tick(object sender, EventArgs e)
        {
            CheckCursorPositionChanged();
        }

        /// <summary>
        /// This method check if cursor position changed.
        /// </summary>
        private void CheckCursorPositionChanged()
        {
            Word.Application currentTemplate = Globals.ThisAddIn.Application;

            if (currentTemplate == null)
            {
                return;
            }

            if (currentTemplate.Selection != null)
            {
                int startIndex = currentTemplate.Selection.Range.Start;

                if (lastStartIndex != startIndex)
                {
                    refreshButton.PerformClick();
                    lastStartIndex = startIndex;
                }
            }
        }
    }
}