﻿using System.Collections.Generic;
using TemplateBuilder.Model.Entities.NodeObjects.Methods;

namespace TemplateBuilder.Model.Entities.NodeObjects
{
    /// <summary>
    /// This class represents attribute in metadata tree view.
    /// </summary>
    public class NodeObjectAttribute : NodeObject
    {
        /// <summary>
        /// List of methods for the current instance.
        /// </summary>
        public List<NodeObjectMethod> Methods { get; set; }

        /// <summary>
        /// Data type of current attribute.
        /// </summary>
        public EDataType DataType { get; private set; }

        /// <summary>
        /// Flag: true - if current instance (attribute) has methods; false - otherwise.
        /// </summary>
        public bool HasMethods
        {
            get
            {
                if (Methods == null || Methods.Count <= 0)
                {
                    return false;
                }

                return true;
            }
        }

        /// <summary>
        /// Language construction of current node object to insert into document template.
        /// </summary>
        public override string TemplateMark
        {
            get
            {
                string mark = NodeObjectFullName;

                if (HasMethods)
                {
                    foreach(NodeObjectMethod currentMethod in Methods)
                    {
                        mark += currentMethod.ToString();
                    }
                }

                return SetSpecialCharsToMark(mark);
            }
        }

        /// <summary>
        /// Constructor of the NodeObjectAttribute class.
        /// </summary>
        /// <param name="nodeObjectFullName">Full name of the node object (attribute).</param>
        /// <param name="dataType">Data type of current attribute.</param>
        /// <param name="label">Label of the current attribute.</param>
        public NodeObjectAttribute(string nodeObjectFullName, EDataType dataType, string label) : base (nodeObjectFullName, label)
        {
            ObjectType = Metadata.ENodeObjectType.Attribute;
            DataType = dataType;
        }

        /// <summary>
        /// This method adds new method to the current instance (attribute).
        /// </summary>
        /// <param name="methodType">Specific type of attribute method.</param>
        /// <param name="parametersStr">Parameters of the method.</param>
        public void AddMethod(ENodeObjectMethodType methodType, params string[] parametersStr)
        {
            if (Methods == null)
            {
                Methods = new List<NodeObjectMethod>();
            }

            if (ContainsMethod(methodType))
            {
                RemoveMethod(methodType);
            }

            if (methodType.Equals(ENodeObjectMethodType.FormatDate))
            {
                if (parametersStr.Length == 1)
                {
                    Methods.Add(new NodeObjectFormatDateMethod(parametersStr[0]));
                }
                else
                {
                    throw new AttributeException(ErrorType.INVALID_NUM_OF_ATTRS);
                }
            }
        }

        /// <summary>
        /// This method removes method from current instance (attribute) 
        /// according to the method type.
        /// </summary>
        /// <param name="methodType">Specific type of attribute method.</param>
        public void RemoveMethod(ENodeObjectMethodType methodType)
        {
            if (!ContainsMethod(methodType))
            {
                return;
            }

            for (int i = 0; i < Methods.Count; i++)
            {
                NodeObjectMethod currentMethod = Methods[i];

                if (currentMethod.MethodType.Equals(methodType))
                {
                    Methods.RemoveAt(i);
                }
            }
        }

        /// <summary>
        /// This method finds out if the current instance (attribute) 
        /// contains method specified by method type.
        /// </summary>
        /// <param name="methodType">Specific type of attribute method.</param>
        /// <returns>true - if current instance (attribute) contains specified method; false - otherwise.</returns>
        private bool ContainsMethod(ENodeObjectMethodType methodType)
        {
            if (Methods == null )
            {
                return false;
            }

            foreach(NodeObjectMethod currentMethod in Methods)
            {
                if (currentMethod.MethodType.Equals(methodType))
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// This method adds special characters to language construction.
        /// </summary>
        /// <param name="mark">Language construction without special characters.</param>
        /// <returns>Language construction with special characters.</returns>
        private string SetSpecialCharsToMark(string mark)
        {
            return "${" + mark + "}";
        }
    }
}
