﻿using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json.Schema;
using System.Collections.Generic;
using System.IO;
using TemplateBuilder.Model.Entities;

namespace TemplateBuilder.Model.Services
{
    /// <summary>
    /// This class provides to load metadata from JSON file and validate them.
    /// </summary>
    public class MetadataLoader
    {
        /// <summary>
        /// Root of the metadata.
        /// </summary>
        public MetadataObject Root { get; set; } = null;

        /// <summary>
        /// This method loads metadata from JSON file.
        /// </summary>
        /// <param name="path">Metadata JSON file.</param>
        public void LoadMetadataJson(string path)
        {
            if (!ValidateMetadataJSON(path))
            {
                throw new MetadataException(ErrorType.METADATA_FILE_INVALID_STRUCTURE);
            }

            using (StreamReader r = new StreamReader(path))
            {
                string json = r.ReadToEnd();
                Root = JsonConvert.DeserializeObject<MetadataObject>(json);
            }

        }

        /// <summary>
        /// This method validates metadata JSON file according to defined JSON schema.
        /// </summary>
        /// <param name="path">Metadata JSON file.</param>
        /// <returns>true - if metadata JSON file is valid, false - otherwise.</returns>
        private bool ValidateMetadataJSON(string path)
        {
            string data = File.ReadAllText(path);
            string schema = Properties.Resources.metadata_schema.ToString();
            JObject model;
            JSchema json_schema;

            try
            {
                model = JObject.Parse(data);
                json_schema = JSchema.Parse(schema);
                IList<string> messages;
                return model.IsValid(json_schema, out messages);
            }
            catch (JsonReaderException e)
            {
                throw new MetadataException(ErrorType.METADATA_FILE_INVALID_STRUCTURE + " " + e.Message);
            }
        }

    }
}
