﻿using System.Windows.Forms;
using TemplateBuilder.Model.Entities;
using TemplateBuilder.Model.Entities.Metadata;
using TemplateBuilder.Model.Entities.NodeObjects;

namespace TemplateBuilder.Model.Services
{
    /// <summary>
    /// This class provides to work with nodes of Metadata Tree View.
    /// </summary>
    public class NodeManager
    {
        /// <summary>
        /// Node of the TreeView.
        /// </summary>
        public TreeNode Node { get; set; }

        /// <summary>
        /// This attribute contains current node object.
        /// </summary>
        private NodeObject nodeObject = null;

        /// <summary>
        /// This attribute contains full name of node object.
        /// </summary>
        private string nodeObjectFullName = null;

        /// <summary>
        /// Property of the 'nodeObjectFullName' attribute.
        /// </summary>
        public string NodeObjectFullName
        {
            get
            {
                if (nodeObjectFullName == null)
                {
                    nodeObjectFullName = ResolveFullName();
                }

                return nodeObjectFullName;
            }
        }

        /// <summary>
        /// Property of the 'nodeObject' attribute.
        /// </summary>
        public NodeObject NodeObject
        { 
            get
            { 
                if (nodeObject == null)
                {
                    nodeObject = ResolveNodeObject();
                }

                return nodeObject;
            }
        }

        /// <summary>
        /// Language construction of current node object to insert into document template.
        /// </summary>
        public string TemplateMark
        {
            get
            {
                return NodeObject.TemplateMark;
            }
        }

        /// <summary>
        /// Constructor of the NodeManager class.
        /// </summary>
        /// <param name="node">Node of the TreeView.</param>
        public NodeManager(TreeNode node)
        {
            Node = node;
        }

        /// <summary>
        /// This method resolves type of the node object.
        /// </summary>
        /// <returns>Current node object.</returns>
        private NodeObject ResolveNodeObject()
        {
            object nodeTag = ((MetadataParams)Node.Tag).MetadataClass;

            if (nodeTag is MetadataAttribute attribute)
            {
                return new NodeObjectAttribute(NodeObjectFullName, attribute.DataType, attribute.Label);
            }
            else if (nodeTag is MetadataObject element)
            {
                element.IsExpanded = !element.IsExpanded;

                return new NodeObjectElement(NodeObjectFullName, element.IsArray, element.Label);
            }
            else
            {
                throw new NodeObjectException(ErrorType.TREE_NODE_NOT_VALID);
            }
        }

        /// <summary>
        /// This method resolves full name of the node object.
        /// </summary>
        /// <returns>Full name of the node object.</returns>
        private string ResolveFullName()
        {
            string fullName = string.Empty;

            MetadataParams metadataParams = (MetadataParams)Node.Tag;

            TreeNode currentNode = Node;

            while (currentNode != null)
            {
                fullName = "." + currentNode.Name + fullName;

                currentNode = currentNode.Parent;
            }

            fullName = fullName.Remove(0, 1);

            if (metadataParams.ControlVariableId != null && metadataParams.ControlArray != null)
            {
                string[] fullNameSplitted = fullName.Split('.');

                fullNameSplitted[0] = metadataParams.ControlVariableId;

                fullName = string.Join(".", fullNameSplitted);
            }

            return fullName;
        }
    }
}
