﻿using System;
using System.Drawing;
using System.Windows.Forms;
using TemplateBuilder.Model.Entities;
using TemplateBuilder.Model.Entities.Metadata;
using TemplateBuilder.Model.Utils;

namespace TemplateBuilder.Model.Services
{
    /// <summary>
    /// This class provides to work with Metadata Tree View.
    /// </summary>
    public class TreeViewManager
    {
        /// <summary>
        /// Instance of TreeView.
        /// </summary>
        private TreeView _treeView;

        /// <summary>
        /// Constructor of the TreeViewManager class.
        /// </summary>
        /// <param name="treeView">Instance of TreeView.</param>
        public TreeViewManager(TreeView treeView)
        {
            _treeView = treeView;

            ClearTreeView();
        }

        /// <summary>
        /// This method updates the current tree view.
        /// </summary>
        /// <param name="node">Node of MetadataObject.</param>
        /// <param name="controlVariableId">Identifier of the controlling variable.</param>
        /// <param name="controlArray">Identifier of the controlling array.</param>
        public void UpdateTreeView(MetadataObject node, string controlVariableId, string controlArray)
        {
            _treeView.BeginUpdate();
            TreeNode treeRoot = _treeView.Nodes.Add(node.Name, node.Label);
            treeRoot.Tag = new MetadataParams(node);

            try
            {
                MetadataExecutor.ExecuteNode(treeRoot, node, controlVariableId, controlArray);
            }
            catch(Exception ex)
            {
                string infoMessage = "Updating metadata tree view failed.\nError:\n" + ex.Message;
                MessageBox.Show(infoMessage, "Error: Updating metadata treeview", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            _treeView.EndUpdate();
        }

        /// <summary>
        /// This method clears up the tree view.
        /// </summary>
        private void ClearTreeView()
        {
            _treeView.Nodes.Clear();

            _treeView.ImageList = SetImageList();
            _treeView.ItemHeight = 20;
        }

        /// <summary>
        /// This method provides to convert MSO images and sets them to image list for the tree view.
        /// </summary>
        /// <returns>Current list of images.</returns>
        private static ImageList SetImageList()
        {
            ImageList imageList = new ImageList();

            Image nodeImage = ImageConvertor.Convert(Globals.ThisAddIn.Application.CommandBars.GetImageMso("WebSiteMapHierarchView", 16, 16));
            Image dateTimeImage = ImageConvertor.Convert(Globals.ThisAddIn.Application.CommandBars.GetImageMso("DataTypeMediumDate", 16, 16));
            Image numberImage = ImageConvertor.Convert(Globals.ThisAddIn.Application.CommandBars.GetImageMso("DataTypeNumber", 28, 28));
            Image textImage = ImageConvertor.Convert(Globals.ThisAddIn.Application.CommandBars.GetImageMso("DataTypeRichText", 26, 26));
            Image directiveImage = ImageConvertor.Convert(Globals.ThisAddIn.Application.CommandBars.GetImageMso("ToolboxGallery", 20, 20));
            Image defaultImage = ImageConvertor.Convert(Globals.ThisAddIn.Application.CommandBars.GetImageMso("WebSiteMapHierarchView", 24, 24));

            imageList.Images.Add(Constants.NODE_IMAGE_KEY, nodeImage);
            imageList.Images.Add(Constants.DATE_TIME_IMAGE_KEY, dateTimeImage);
            imageList.Images.Add(Constants.NUMBER_IMAGE_KEY, numberImage);
            imageList.Images.Add(Constants.TEXT_IMAGE_KEY, textImage);
            imageList.Images.Add(Constants.DIRECTIVE_IMAGE_KEY, directiveImage);
            imageList.Images.Add(Constants.DEFAULT_IMAGE_KEY, defaultImage);

            return imageList;
        }

        /// <summary>
        /// This method resolves if the current tree view is according to input metadata node.
        /// </summary>
        /// <param name="node">Input metadata node.</param>
        /// <returns>true - if the current tree view is according to input metadata node; false - otherwise.</returns>
        public bool IsSameTree(MetadataObject node)
        {
            TreeNode treeRoot = GetTreeRoot();

            if (treeRoot.Name.Equals(node.Name))
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// This method gets root of the tree view.
        /// </summary>
        /// <returns>Root of the tree view.</returns>
        private TreeNode GetTreeRoot()
        {
            TreeNode node = _treeView.Nodes[0];

            while (node.Parent != null)
            {
                node = node.Parent;
            }

            return node;
        }
    }
}
