﻿using Microsoft.Office.Tools.Ribbon;
using System;
using System.Windows.Forms;
using TemplateBuilder.Forms;
using TemplateBuilder.Model.Services;

namespace TemplateBuilder
{
    /// <summary>
    /// This partial class serves actions for TemplateRibbon. 
    /// </summary>
    public partial class TemplateRibbon
    {
        /// <summary>
        /// Instance of global WinForms dialogs.
        /// </summary>
        private Dialogs _dialogs = null;

        /// <summary>
        /// Instance of MetadataLoader to load metadata file.
        /// </summary>
        private MetadataLoader _metadataLoader = null;

        /// <summary>
        /// WinForms Task Pane to view Metadata Tree.
        /// </summary>
        private TemplateTaskPane _templateTaskPane = null;

        /// <summary>
        /// This method initializes main attributes of the TemplateRibbon.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void TemplateRibbon_Load(object sender, RibbonUIEventArgs e)
        {
            _dialogs = Globals.ThisAddIn.Dialogs;
            _metadataLoader = Globals.ThisAddIn.MetadataLoader;
            _templateTaskPane = Globals.ThisAddIn.TemplateTaskPane;
        }

        /// <summary>
        /// This method serves action on load button click in TemplateRibbon.
        /// It provides to open metadata choose file dialog, to load metadata
        /// and to initialize metadata tree view.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void LoadButton_Click(object sender, RibbonControlEventArgs e)
        {
            string filePath = _dialogs.MetadataFileChooseDialog();

            if (filePath != null && filePath.Length > 0)
            {
                try
                {
                    _metadataLoader.LoadMetadataJson(filePath);
                    _templateTaskPane.TreeViewInit(_metadataLoader.Root);
                }
                catch (Exception ex)
                {
                    string infoMessage = "Loading metadata JSON file failed.\nError:\n" + ex.Message;
                    MessageBox.Show(infoMessage, "Error: Metadata JSON file", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }

            }
        }

        /// <summary>
        /// This method serves action on show tree button click in TemplateRibbon.
        /// It show metadata tree view.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void ShowTreeButton_Click(object sender, RibbonControlEventArgs e)
        {
            Globals.ThisAddIn.CustomTaskPaneTemplate.Visible = true;
        }

        /// <summary>
        /// This method serves action on hide tree button click in TemplateRibbon.
        /// It hides metadata tree view.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void HideTreeButton_Click(object sender, RibbonControlEventArgs e)
        {
            Globals.ThisAddIn.CustomTaskPaneTemplate.Visible = false;
        }

        /// <summary>
        /// This method serves action on methods gallery click in TemplateRibbon.
        /// It expands the gallery and shows available methods to use in document template.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void MethodsGallery_Click(object sender, RibbonControlEventArgs e)
        {
            RibbonDropDownItem item = methodsGallery.SelectedItem;

            string itemLabel = item.Label;

            if (itemLabel.Equals("Current Date"))
            {
                CurrentDateForm currentDate = new CurrentDateForm();
                currentDate.Show();
            }
        }

        /// <summary>
        /// This method serves action on directives gallery click in TemplateRibbon.
        /// It expands the gallery and shows available directives to use in document template.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void DirectivesGallery_Click(object sender, RibbonControlEventArgs e)
        {
            RibbonDropDownItem item = directivesGallery.SelectedItem;

            string itemLabel = item.Label;

            if (itemLabel.Equals("Include Template"))
            {
                IncludeForm include = new IncludeForm();
                include.Show();
            }
        }

        /// <summary>
        /// This method serves action on generate button click in TemplateRibbon.
        /// It opens generate form.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void GenerateButton_Click(object sender, RibbonControlEventArgs e)
        {
            GenerateForm generateForm = new GenerateForm();
            generateForm.Show();
        }

        /// <summary>
        /// This method serves action on show errors button click in TemplateRibbon.
        /// It opens ErrorTaskPane with syntax errors.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void ShowErrorsButton_Click(object sender, RibbonControlEventArgs e)
        {
            Globals.ThisAddIn.CustomTaskPaneError.Visible = true;
        }

        /// <summary>
        /// This method serves action on hide errors button click in TemplateRibbon.
        /// It hides ErrorTaskPane with syntax errors.
        /// </summary>
        /// <param name="sender">Sender.</param>
        /// <param name="e">Event arguments.</param>
        private void HideErrorsButton_Click(object sender, RibbonControlEventArgs e)
        {
            Globals.ThisAddIn.CustomTaskPaneError.Visible = false;
        }
    }
}
